import re
import numpy as np
import pkg_resources
import yaml
import inspect

import gherkan.utils.constants as c

class SignalCondition():
    def __init__(self, string, kind, negated=False):
        self.kind = kind
        self.string = string
        self.variable = None
        self.value = None
        self.variableTr = None
        self.templateID = None
        self.variableNLP = None
        self.variableNLPCZ = None
        self.variableNL_full = None
        self.variableNL_full_cs = None   

        if self.kind == c.EQUALITY or kind == c.INEQUALITY:
            match = re.search(r'(?P<variable>\w+)\s*[=!]+\s*(?P<value>\w+)', self.string)
            if negated:
                if self.kind == c.EQUALITY:
                    self.kind = c.INEQUALITY
                else:
                    self.kind = c.EQUALITY
        elif self.kind == c.BOOL:
            match = re.search(r'(?P<variable>\w+)', self.string)
            self.value = 1 if not negated else 0
        elif self.kind == c.EDGE:
            match = re.search(r'edge\((?P<variable>\w+)\s*,\s*(?P<value>\w+)', self.string)
        elif self.kind == c.FORCE:
            match = re.search(r'force\((?P<variable>\w+)\s*,\s*(?P<value>\w+)', self.string)
        elif self.kind == c.UNFORCE:
            match = re.search(r'unforce\((?P<variable>\w+)', self.string)
        else:
            raise Exception("Error while parsing a signal condition (whaaat?). Related string: '{}'. Unexpected condition kind: '{}'.".format(self.string, self.kind))
        if match:
            self.variable = match.group("variable")
            if "value" in match.groupdict().keys():
                self.value = match.group("value")
        else:
            raise Exception("Error while parsing a signal condition. Phrase matching failed!\nRelated string: '{}'\nCondition kind: '{}'".format(self.string, self.kind))


class NLCondition():
    def __init__(self, strFullEn = None, strFullCs = None, strEng = None, strCZ = None):
        self.kind = None
        self.string = strFullEn
        self.variable = None
        self.value = None
        self.variableTr = None
        self.templateID = None
        self.variableNL_full = strFullEn
        self.variableNL_full_cs = strFullCs                
        self.variableNLP = strEng
        self.variableNLPCZ = strCZ


# TODO defined elsewhere, remove
# class SignalStatementTreeNode():
#
#     def __init__(self, kind=None, data=None):
#         self.__kind = kind
#         self.__data = data
#         self.__parent = None
#
#     def setParent(self, parent):
#         self.__parent = parent
#
#     @property
#     def kind(self):
#         return self.__kind
#
#     @property
#     def parent(self):
#         return self.__parent
#
#     @property
#     def data(self):
#         return self.__data
#
#     def getIncompleteParent(self):
#         raise NotImplementedError()
#
#
#
# class SignalStatementTreeOperatorNode(SignalStatementTreeNode):
#     AND = 'and'
#     OR = 'or'
#
#     def __init__(self, kind):
#         super().__init__(kind.lower())
#         self.__lchild = None  # type: SignalStatementTreeNode
#         self.__rchild = None  # type: SignalStatementTreeNode
#         self.templateID = None
#         self.variableNLP = None
#         self.variableNL_full = None
#         self.variableNLPCZ = None
#         self.variableNL_full_cs = None
#
#     @property
#     def lchild(self):
#         return self.__lchild
#
#     @property
#     def rchild(self):
#         return self.__rchild
#
#     @property
#     def children(self):
#         return [self.lchild, self.rchild]
#
#     @property
#     def full(self):
#         return (self.lchild is not None) and (self.rchild is not None)
#
#     def getIncompleteParent(self):
#         if self.full:
#             if self.parent:
#                 return self.parent.getIncompleteParent()
#             else:
#                 return None
#         else:
#             return self
#
#     def attachChild(self, child: SignalStatementTreeNode):
#         """
#         Returns True if the node is "full" (i.e. both children are assigned)
#         """
#         child.setParent(self)
#         if self.lchild:
#             self.__rchild = child
#             return True
#         else:
#             self.__lchild = child
#             return False
#
#     def __str__(self):
#         return 'Operator(type: {})'.format(self.kind.upper())


# class SignalStatementTreeOperandNode(SignalStatementTreeNode):
#
#     def __init__(self, data: SignalCondition):
#         super().__init__(data.kind, data)
#
#     def getIncompleteParent(self):
#         self.parent.getIncompleteParent()
#
#     def __str__(self):
#         return 'Operand(type: {}, variable: {}, value: {}, variableNLP: {}, variableNL_full: {})'.format(self.kind.upper() if self.kind else 'None', self.data.variable, self.data.value, self.data.variableNLP, self.data.variableNL_full)
#
#     @property
#     def variable(self):
#         return self.data.variable
#
#     @property
#     def kind(self):
#         return self.data.kind
#
#     @property
#     def variableNLP(self):
#         return self.data.variableNLP
#
#     @property
#     def variableNL_full(self):
#         return self.data.variableNL_full
#
#     @property
#     def value(self):
#         return self.data.value
#
#
# class SignalStatementTree():
#     dash, line, left, right = '\u2500', '\u2502', '\u251c\u2500\u2500 ', '\u2514\u2500\u2500 '
#
#     def __init__(self, root: SignalStatementTreeNode, string: str):
#         self.__root = root
#         self.__string = string
#
#     @property
#     def root(self):
#         return self.__root
#
#     @property
#     def string(self):
#         return self.__string
#
#     def __printNode(self, node: SignalStatementTreeNode, leftChild, levelsOpened):
#         string = ''.join([SignalStatementTree.line + "\t" if lo else "\t" for lo in levelsOpened[:-1]])
#         if leftChild:
#             if node:
#                 string += SignalStatementTree.left + str(str(node) if bool(node) else "ERROR!!!") + '\n'
#         else:
#             if node:
#                 string += SignalStatementTree.right + str(str(node) if bool(node) else "ERROR!!!") + '\n'
#
#         if type(node) is SignalStatementTreeOperatorNode:
#             string += self.__printNode(node.lchild, True, levelsOpened + [True])
#             string += self.__printNode(node.rchild, False, levelsOpened + [False])
#         return string
#
#     def __str__(self):
#         levelsOpened = [False]
#
#         node = self.root
#         string = str(node) + '\n'
#         if type(node) is SignalStatementTreeOperatorNode:
#             string += self.__printNode(node.lchild, True, levelsOpened + [True])
#             string += self.__printNode(node.rchild, False, levelsOpened + [False])
#         return string
#

class EvalFunction():
    """
    Place holder for functions to be evaluated on the SignalStatementTree.
    Contains additional functionality to enable creation of more complex templates.

    Attributes
    ----------
    func : Function
        The function to be evaluated on SignalStatementTreeNode objects (altough the implementation does not enforce any class)

    succeeded : bool
        True if the function was evaluated to True sufficient number of times

    successes : int
        The number of times when the function was evaluated successfully
    """

    def __init__(self, func, minSuccesses=1, maxSuccesses=None, failFast=False):
        """
        Parameters
        ----------
        func : Function
            The function to be evaluated on SignalStatementTreeNode objects

        minSuccesses : int
            The minimum number of times the function needs to evaluate to True
            to declare overall success. Use this in combination with maxSuccesses
            to have a range of possible nodes that comply to the requirements
            enforced by the function.

        maxSuccesses : int
            The maximum number of cases the function can be evaluated to True.
            After the maxSuccesses has been reached, all following evaluations
            will fail, even if they would succeed otherwise. Set this value
            to a very large number to achieve unlimited number of successful
            evaluations.

        failFast : bool
            If True, the first case when the function is evaluated to False
            will result in complete failure of this condition. I.e. all future
            evaluations will only fail and overall success will be False
            no matter what. Use this to enforce a condition on all tested nodes.
            failFast will also cause permanent failure if number of maxSuccesses
            is exceeded. This can be used to limit the number of nodes in a subtree.
            E.g., EvalFunction(lambda x: True, maxSuccesses=5, failFast=True)
            Will generate a function that will make sure that there are at most 5
            nodes in the subtree.
        """
        self.__func = func
        self.__succeeded = None
        self.__successes = 0
        self.__minSuccesses = minSuccesses
        self.__failFast = failFast
        self.__valid = True
        if self.__failFast:
            self.__maxSuccesses = maxSuccesses if maxSuccesses else np.iinfo(np.int).max
        else:
            self.__maxSuccesses = maxSuccesses if maxSuccesses else minSuccesses

    def invalidate(self):
        self.__valid = False

    def __bool__(self):
        if self.__succeeded is None:
            if self.successes >= self.minSuccesses:
                self.__succeeded = True
                return True
            else:
                return False
        elif self.__succeeded is True:
            return True
        elif self.__succeeded is False:
            return False

    def __call__(self, value):
        """Calls the associated function and evaluates the result depending on other conditions as well.
        If maxSuccesses has been exceeded, the call will fail no matter the actual result
        (the function won't even be evaluated). It the number of successes (evaluations/calls resulting in True)
        is smaller than maxSuccesses, the function, associated with this placeholder, is called with the provided value.
        If the call returns True, number of successes is increased. The success, i.e. the result of the call is returned.
        If failFast is set to True, any call that will result in return value equal to False will also fail the success
        of the entire placeholder (EvalFunction instance)

        Parameters
        ----------
        value : any
            Function parameters

        Returns
        -------
        bool
            The success of the call (depending on other parameters)
        """
        if self.__succeeded is not None and not self.__succeeded:
            return False
        if self.successes >= self.maxSuccesses:
            if self.failFast and self.func(value):
                self.__succeeded = False
            return False
        success = self.func(value)
        if success:
            self.__successes += 1
        elif self.failFast:
            self.__succeeded = False
        return success

    @property
    def successes(self):
        return self.__successes

    @property
    def succeeded(self):
        return bool(self)

    @property
    def func(self):
        return self.__func

    @property
    def minSuccesses(self):
        return self.__minSuccesses

    @property
    def maxSuccesses(self):
        return self.__maxSuccesses

    @property
    def valid(self):
        return self.__valid

    @property
    def failFast(self):
        return self.__failFast

    def __repr__(self):
        sig = inspect.signature(self.func)

        def displayParam(p, d):
            return "\t\t\t{}:\n{}".format(p, "\n".join(["\t\t\t\t{}".format(item) for item in d])) if type(d) is list else "\t\t\t{}: {}".format(p, d)

        return "\n".join([
            ">>> EvalFunction object <<<",
            "\tNumber of successes = {}".format(self.successes),
            "\tMinimum number of successes = {}".format(self.minSuccesses),
            "\tMaximum number of successes = {}".format(self.maxSuccesses),
            "\tSucceeded = {}".format(self.succeeded),
            "\tAssociated function:",
            "\t\tSignature: {}".format(str(sig)),
            "\t\tParameters: ",
            "\n".join(["\t\t\t" + param if sig.parameters[param].default is inspect._empty else displayParam(param, sig.parameters[param].default) for param in sig.parameters]),
            "\t\tSource code: \n" + inspect.getsource(self.func).strip()
        ])


class EvalFunctionList():

        def __init__(self, functionList):
            """
            Helper function to generate so called structure list from a list of functions.
            Changed to generate a list of EvalFunction objects
            For example:
            [lambda x: x + 2, lambda x: x * 3] -> [EvalFunction(lambda x: x + 2), EvalFunction(lambda x: x * 3)]
            """
            self.evalList = []
            for item in functionList:
                if callable(item):  # just the function; default arguments will be used
                    evalList.append(EvalFunction(item))
                elif type(item) is tuple:  # a list of positional arguments
                    evalList.append((EvalFunction(*item)))
                elif type(item) is dict:  # dictionary with kwargs
                    evalList.append((EvalFunction(**item)))
                else:
                    raise Exception("Unknown object type encountered while constructing structure list.")

# TODO make iterable and validable
# TODO incorporate additional methods into SignalStatementTree

# class SignalTemplateTree(SignalStatementTree):
#
#     def __init__(self, signalTree: SignalStatementTree):
#         super().__init__(signalTree.root, signalTree.string)
#
#     def __isTwig(self, node: SignalStatementTreeNode):
#         """
#         Checks if a node is an operator node but has only leaves (i.e. no consecutive operator nodes == no further branching)
#         """
#         if type(node) is SignalStatementTreeOperatorNode:
#             return (type(node.lchild) is SignalStatementTreeOperandNode) and (type(node.rchild) is SignalStatementTreeOperandNode)
#         else:
#             return False
#
#     def __recursiveBoolEval(self, node: SignalStatementTreeNode, func):
#         """
#         Recursively attempts to evaluate the function on all subsequent nodes.
#         Returns True if all evaluations were successful. That is, if the function
#         is holds true for all nodes in the subtree.
#         """
#         result = func(node)
#         if type(node) is SignalStatementTreeOperandNode:
#             return result
#         elif type(node) is SignalStatementTreeOperatorNode:
#             return result and self.__recursiveBoolEval(node.lchild, func) and self.__recursiveBoolEval(node.rchild, func)
#         else:
#             # Unknown node type
#             return False
#
#     def __recursiveStructureEval(self, node: SignalStatementTreeNode, structureList: list):
#         """
#         Recursively enters nodes and attempts to evalute functions in the structureList list.
#         If one of the evaluations is successful it then attempts to continue with evaluations
#         on the node's children.
#
#         Old version:
#         "However, the successfully evaluated function is omitted
#         from further evaluations. That is, each function must succeed exactly once!"
#
#         Newer version:
#         "The number of successful evaluations depends on the specific conditions
#         imposed by the EvalFunction object"
#
#         Functions are evaluated in a greedy manner. The order of the functions
#         should therefore be considered - e.g., if some function A can succeed on more nodes
#         than other function B in the list (i.e., domain of B(x) = True is a subset of A(x) = True),
#         B should precede A in the list. Otherwise function A might take node that should
#         normally be captured by B.
#         """
#         if len(structureList) == 0:  # either empty list has been provided or some node emptied it (see !success condition below)
#             return False
#         success = False
#         allTrue = True
#         for i, evalFunc in enumerate(structureList):
#             if not evalFunc.valid:
#                 # TODO how to handle allTrue?????
#                 continue
#             # Run the eval function on the node.
#             # The boolean value of the function object is assigned to the allTrue variable to check if the structure is completed
#             # the internal success of the function is handled by the EvalFunction object
#             if not success:  # if no function succeeded so far
#                 # call the function with the current node as a parameter
#                 success = evalFunc(node)  # if the call returned True, i.e. the node comforms to the condition imposed by the function
#                 # capture the first successful evaluation and do not call any other function
#                 # the list has to be iterated through though to check if all function succeeded already
#             allTrue &= bool(evalFunc)
#             if success and not allTrue:
#                 break
#         if allTrue:  # structureList has been completed
#             return True
#         if success:
#             if type(node) is SignalStatementTreeOperatorNode:
#                 # call function on both children
#                 self.__recursiveStructureEval(node.lchild, structureList)
#                 self.__recursiveStructureEval(node.rchild, structureList)
#                 if len(structureList) > 0:  # if the list hasn't been emptied yet - the structure check did not fail so far
#                     return all([bool(e) for e in structureList])  # return the current completion status of the structure check
#             return False
#         else:
#             # if none of the functions in the list succeeded (i.e. the node is not part of the structure), then the structure check failed
#             # that is, this was not the structure we were looking for...
#             for evalFunc in structureList:
#                 evalFunc.invalidate()
#             return False
#
#     def __recursiveSubTreeSearch(self, node: SignalStatementTreeNode, func):
#         """
#         Returns all subtrees (returns their respective roots in a list, to be more exact),
#         for which the function evalutes to True.
#
#         Parameters
#         ----------
#         node : SignalStatementTreeNode
#             A node from the tree on which the function should be evaluted
#
#         func: Function
#             Function to be evaluated
#
#         Returns
#         -------
#         list OR bool:
#             List of roots of subtrees for which the function was evaluted to True.
#             Returns just boolean True if this node evaluted to True. Otherwise,
#             if this node evaluted to False but its child evaluted to True,
#             a list containing this child will be returned.
#         """
#         if func(node):
#             return node
#         else:
#             if type(node) is not SignalStatementTreeOperatorNode:
#                 return False
#             else:
#                 rootList = []
#                 # Evaluate the function on children and iterate over the results
#                 for res in [self.__recursiveSubTreeSearch(node.lchild, func), self.__recursiveSubTreeSearch(node.rchild, func)]:
#                     if not res:
#                         continue
#                     if type(res) is list:
#                         rootList += res
#                     elif type(res) is SignalStatementTreeNode:
#                         rootList += [res]
#
#                 return rootList
#
#     def __genNextTemplate(self):
#         """
#         Generates function that should search for templates in the tree.
#
#         Raises:
#             Exception -- Raises an exception if the ID of the requested template
#             does not exist. That is, function for the template is not specified, yet.
#         """
#         with pkg_resources.resource_stream("gherkan", '/'.join(["containers", "templates_dic.yaml"])) as f:
#             templateDict = yaml.load(f)
#
#         for templateID, templateSignature in templateDict.items():
#             if "SignalTemplate" in templateSignature:  # find signal template that has the property called "SignalTemplate"
#                 signalTemplateSignature = templateSignature["SignalTemplate"]
#                 template = None
#                 # find what kind of template it has (currently only structure is supported)
#                 if type(signalTemplateSignature) is list or "structure" in signalTemplateSignature:
#                     if type(signalTemplateSignature) is list:
#                         templateStructure = signalTemplateSignature
#                     elif type(signalTemplateSignature) is dict:
#                         templateStructure = signalTemplateSignature["structure"]
#                     structureList = []
#                     for item in templateStructure:
#                         if type(item) is str:
#                             structureList.append(lambda nd, ndkind=item: type(nd) is SignalStatementTreeOperatorNode and nd.kind == ndkind)
#                         elif type(item) is dict:
#                             lamlist = []
#                             kind, paramdict = item.popitem()
#                             for key, value in paramdict.items():
#                                 lamlist.append(lambda nd, var=key, val=value:  val.lower() in getattr(nd.data, var).lower())
#                             structureList.append(lambda nd, ndkind=kind, paramCheckList=lamlist: type(nd) is SignalStatementTreeOperandNode and nd.kind == ndkind and all([check(nd) for check in paramCheckList]))
#                         else:
#                             raise Exception("Unknown type in signalTemplateSignature.\n{}".format(signalTemplateSignature))
#
#                     def template(node: SignalStatementTreeNode, structure=self.__genStructureFromFunctionList(structureList)):
#                         def searchFunction(node: SignalStatementTreeNode):
#                             return self.__recursiveStructureEval(node, structure)
#                         return self.__recursiveSubTreeSearch(node, searchFunction)
#
#                 elif "condition" in signalTemplateSignature:
#                     pass
#                 elif "script" in signalTemplateSignature:
#                     pass
#
#                 yield((templateID, template))
#
#     def parseTemplates(self):
#         for templateID, template in self.__genNextTemplate():
#             if not template:
#                 raise Exception("Error while generating template from SignalTemplateTree!")
#             tempResult = template(self.root)
#             if tempResult:
#                 print(tempResult)
#                 if type(tempResult) is not list:
#                     tempList = [tempResult]
#                 for subtreeRoot in tempList:
#                     subtreeRoot.templateID = templateID




import datetime
import importlib
import os
import re

import yaml

from .raw_text_to_signal import process as nlToSignal
from .signal_to_nl import process as signalToNL
from .speech_to_text import process as speechToText

from flask import send_file


class API_FSA():
    """
    FSA class for tracking state from REST API. State is written into a file
    so it should be persistent even over long time periods.
    """

    # =====>>> PATH variables <<<===== #
    stateFilePath = 'state.fsa'
    logFilePath = 'log.txt'
    configFilePath = 'fsa_config.yaml'

    # This path does not need to be set to the package's root, probably
    rootFolder = importlib.util.find_spec("gherkan").submodule_search_locations[0]

    dataFolder = os.path.join(rootFolder, 'data')
    audioFolder = os.path.join(dataFolder, 'audio')
    # nlTextFolder = os.path.join(dataFolder, 'nl_text')
    # signalsFolder = os.path.join(dataFolder, 'signals')

    robotProgramsPath = os.path.join(rootFolder, 'utils', 'RobotPrograms_{}.json')

    language = 'en'

    # =====>>> STATE constants <<<===== #
    S_OFF = 0
    S_IDLE = 1
    S_RECEIVED_NL_TEXT = 2
    S_CORRECTING_NL_TEXT = 4
    S_FINISHED_PROCESSING_NL_TEXT = 8
    S_RECEIVED_SIGNAL = 16
    S_FINISHED_PROCESSING_SIGNAL = 32
    S_RECEIVED_AUDIO = 64
    S_CORRECTING_AUDIO = 128
    S_FINISHED_PROCESSING_AUDIO = 256
    S_ERROR = 1024

    # =====>>> PARAMETER name constants <<<===== #
    P_LAST_SIGNAL_FILE = "lastSignalFile"
    P_LAST_NL_TEXT_FILE = "lastNLTextFile"
    P_LAST_AUDIO_FILE = "lastAudioFile"
    P_RESPONSE_MODE = "responseMode"

    # =====>>> parameter VALUES <<<===== #
    PV_RESPONSE_FILE = "responseModeFile"
    PV_RESPONSE_JSON = "responseModeJSON"

    # =====>>> OTHER <<<===== #
    basenameExtractorRegex = re.compile(r"(?P<basename>.+?)(?P<rest>(_signals|\.)+.+)")

    # =====>>> Helper STATIC methods <<<===== #
    @staticmethod
    def toBytes(integer):
        # Maybe extend this with a loop but so far 1024 as the limit is ok
        return bytes([(0b11100000000 & integer) >> 8, 0b11111111 & integer])

    @staticmethod
    def fromBytes(byte):
        # Maybe extend this with a loop but so far 1024 as the limit is ok
        return (byte[0] << 8) | byte[1]

    @staticmethod
    def generateDatestring():
        return datetime.datetime.now().strftime('%Y_%b_%d-%H_%M_%S')

    # =====>>> Helper CLASS methods <<<===== #
    @classmethod
    def getState(cls):
        with open(cls.stateFilePath, 'rb') as statefile:
            state = cls.fromBytes(statefile.read())
        return state

    @classmethod
    def setState(cls, state):
        if isinstance(state, int):
            state = cls.toBytes(state)
        with open(cls.stateFilePath, 'wb') as statefile:
            state = statefile.write(state)

    @classmethod
    def writeLog(cls, message):
        with open(cls.logFilePath, 'a') as logFile:
            logFile.writeline(message)

    @classmethod
    def readLog(cls):
        with open(cls.logFilePath, 'rb') as logFile:
            logFile.seek(-2, os.SEEK_END)
            while logFile.read(1) != b'\n':
                logFile.seek(-2, os.SEEK_CUR)
            line = logFile.readline().decode()
        return line

    @classmethod
    def getConfig(cls, param=None):
        with open(cls.configFilePath, 'r') as cfg:
            config = yaml.load(cfg)
            if param is not None:
                try:
                    config = config[param]
                except:
                    config = None
        return config

    @classmethod
    def setConfig(cls, param, value):
        with open(cls.configFilePath, 'r') as cfg:
            config = yaml.load(cfg)
        config[param] = value
        with open(cls.configFilePath, 'w') as cfg:
            yaml.dump(config, cfg)

    @classmethod
    def regenerateConfig(cls):
        with open(cls.configFilePath, 'w') as cfg:
            # TODO: add more defaults to config
            yaml.dump({"flush": False, "responseMode": "responseModeFile"}, cfg)

    # =====>>> API initialization <<<===== #
    @classmethod
    def init_fsa(cls, address, port):
        print("The RESTful API to the Gherkan NL Instruction Processing system is starting...")

        # Check if the root folder is set correctly
        if not os.path.isdir(cls.rootFolder):
            raise Exception("The root folder is set to a non-existent path!\nRoot folder: {}".format(cls.rootFolder))
        if not os.access(cls.rootFolder, os.W_OK):
            raise Exception("Cannot write to the root folder!\nRoot folder: {}".format(cls.rootFolder))
        if not os.access(cls.rootFolder, os.R_OK):
            raise Exception("Cannot read from the root folder!\nRoot folder: {}".format(cls.rootFolder))


        # Check the existence of data folders
        if not os.path.isdir(cls.audioFolder):
            os.makedirs(cls.audioFolder)
        # if not os.path.isdir(cls.nlTextFolder):
        #     os.makedirs(cls.nlTextFolder)
        # if not os.path.isdir(cls.signalsFolder):
        #     os.makedirs(cls.signalsFolder)

        if not os.path.isfile(cls.configFilePath):
            cls.regenerateConfig()

        print("Host address: {}\nPort number: {}".format(address, port))
        cls.setConfig("host", address)
        cls.setConfig("port", port)

        print("Variable checking done, checking state from previous run...")
        # Check for previous sessions
        if os.path.exists(cls.stateFilePath):
            previousState = cls.getState()
            # Check and handle error
            if previousState == cls.S_ERROR:
                if cls.getConfig('flush'):
                    print("Found error state from a previous run.\nError flushing is set to True - flushing the error and reseting the system.")
                    previousState = cls.S_OFF
                    cls.regenerateConfig()
                else:
                    message = cls.readLog()
                    raise Exception("Previous run of the system resulted in an error! Resolve the error and set the fsa_config\nLast log message: {}".format(message))
            if previousState != cls.S_OFF:
                print("Found non-OFF state from previous run, restoring the system with state {}".format(previousState))
                cls.setState(previousState)
            else:
                # Set state to ready/idle
                print("System ready, state set to IDLE.")
                cls.setState(cls.S_IDLE)
        else:
            print("System ready, state set to IDLE.")
            cls.setState(cls.S_IDLE)

    # =====>>> RESTful API Call handling functions <<<===== #
    @classmethod
    def receiveSignal(cls, text):
        """This function handles the receiving of signal text.
        """

        state = cls.getState()
        if state & (cls.S_RECEIVED_AUDIO | cls.S_RECEIVED_NL_TEXT):
            raise Exception("Different data type already received!")

        if state & cls.S_RECEIVED_SIGNAL:
            signalFileMode = 'a'
            fileName = cls.getConfig(cls.P_LAST_SIGNAL_FILE)
        else:
            cls.setState(state | cls.S_RECEIVED_SIGNAL)
            signalFileMode = 'w'
            fileName = os.path.join(cls.dataFolder, '_'.join([cls.generateDatestring(), 'signals.feature']))
            cls.setConfig(cls.P_LAST_SIGNAL_FILE, fileName)

        with open(fileName, signalFileMode) as sigFile:
            sigFile.writelines(text)

        # Signal text can be iteratively written into the file
        # When all signals were written, the user can request NL file

    @classmethod
    def requestSignal(cls):
        """This function return a composed signal file derived from NL text, if it was provided
        """
        state = cls.getState()
        if state & cls.S_RECEIVED_SIGNAL:  # Signals were sent and requested back (e.g. to check the state of the file)
            return send_file(cls.getConfig(cls.P_LAST_SIGNAL_FILE))
        elif not (state & cls.S_FINISHED_PROCESSING_NL_TEXT):  # Check if NL text was sent and processed
            raise Exception("NL text was not provided, cannot return signal file!")

        signalFilePath = cls.getConfig(cls.P_LAST_SIGNAL_FILE)

        responseMode = cls.getConfig(cls.P_RESPONSE_MODE)
        if responseMode == cls.PV_RESPONSE_FILE:
            response = send_file(signalFilePath)
        elif responseMode == cls.PV_RESPONSE_JSON:
            with open(signalFilePath, "r") as signalFile:
                signalText = signalFile.read()

            # TODO: Split file into response

            response = {
                    "language": "en<OR>cs",
                    "background": "background text",
                    "description": "This feature is not implemented, yet! Use RESPONSE_MODE: FILE instead.",
                    "scenarios": signalText
            }

        cls.setState(state & ~cls.S_FINISHED_PROCESSING_NL_TEXT)
        return response

    @classmethod
    def requestNLText(cls):
        state = cls.getState()
        if not (state & cls.S_RECEIVED_SIGNAL):
            raise Exception("Signal file was not provided, cannot return NL text!")

        signalFilePath = cls.getConfig(cls.P_LAST_SIGNAL_FILE)

        path, filename = os.path.split(signalFilePath)
        outputFile = os.path.join(path, cls.basenameExtractorRegex.sub(r"\g<basename>.feature", filename))
        
        # print("SignalToNL: ", signalFilePath)
        signalToNL(signalFilePath)

        cls.setState(state & ~cls.S_RECEIVED_SIGNAL)

        responseMode = cls.getConfig(cls.P_RESPONSE_MODE)
        if responseMode == cls.PV_RESPONSE_FILE:
            response = send_file(outputFile)
        elif responseMode == cls.PV_RESPONSE_JSON:
            with open(outputFile, "r") as nlTextFile:
                nlText = nlTextFile.read()

            response = {
                "language":"en<OR>cz",
                "description": "This feature is not implemented, yet! Use RESPONSE_MODE: FILE instead.",
                "background": "background/context text",
                "scenarios": nlText
            }
        else:
            response = {"Message": "Request mode is set to an incorrect value!"}
        return response

    @classmethod
    def receiveNLText(cls, data):
        state = cls.getState()
        if state & (cls.S_RECEIVED_SIGNAL | cls.S_RECEIVED_AUDIO):
            raise Exception("Different data type already received!")

        cls.setState(state | cls.S_RECEIVED_NL_TEXT)

        basepath = os.path.join(cls.dataFolder, cls.generateDatestring())
        nlFilePath = f"{basepath}.feature"
        signalFilePath = f"{basepath}_signals.feature"

        # print("nlToSignal: ", basepath, data)
        nlToSignal(basepath, data)

        cls.setState((state & ~cls.S_RECEIVED_NL_TEXT) | cls.S_FINISHED_PROCESSING_NL_TEXT)
        cls.setConfig(cls.P_LAST_NL_TEXT_FILE, nlFilePath)
        cls.setConfig(cls.P_LAST_SIGNAL_FILE, signalFilePath)

        # TODO: return possible errors
        return {
            "info": "Processing done",
            "lines": [],
            "error_lines": [],
            "error_hints": [],
            "errors": False
        }

    @classmethod
    def receiveAudio(cls, audioPath, language):
        state = cls.getState()
        if state & (cls.S_RECEIVED_SIGNAL | cls.S_RECEIVED_NL_TEXT):
            raise Exception("Different data type already received!")

        if language == "en":
            lang = "en-US"
        elif language == "en":
            lang = "cs-CZ"
        transcriptPath = '.'.join([audioPath[:audioPath.find(".")], ".txt"])

        # print("speechToText: ", audioPath, lang)
        speechToText(audioPath, lang)

        return send_file(transcriptPath)

    @classmethod
    def requestRobotPrograms(cls, language):
        if not language:
            language = cls.language

        with open(cls.robotProgramsPath.format(language), 'r') as robotProgramFile:
            robotPrograms = robotProgramFile.read()

        return robotPrograms

    @classmethod
    def receiveRobotPrograms(cls, programs, language):
        if not language:
            language = cls.language

        with open(cls.robotProgramsPath.format(language), 'w') as robotProgramFile:
            robotProgramFile.writelines(programs)

    # =====>>> RESTful API Helper functions <<<===== #
    @classmethod
    def canRequestSignal(cls):
        """Checks whether it is possible to request a signal file.
        """
        state = cls.getState()
        return state & cls.S_FINISHED_PROCESSING_NL_TEXT

    @classmethod
    def canRequestNLText(cls):
        """Checks whether it is possible to request a NL text file.
        """
        state = cls.getState()
        return state & cls.S_RECEIVED_SIGNAL



# coding: utf-8
"""
Class FillActions turns the filled object SignalPhrase into a natural language phrase, which is then returned
as the filled fields niceStrEN and niceStrCZ in the SignalPhrase.
"""
import yaml, re, json
import os
from termcolor import colored
from gherkan.utils.word_morpho import Finetuner
import gherkan.utils.constants as c

CZ_PROGRAM_DICT_DIR = os.path.join(c.PROJECT_ROOT_DIR, "gherkan/utils", "RobotPrograms_cz.json")
EN_PROGRAM_DICT_DIR= os.path.join(c.PROJECT_ROOT_DIR, "gherkan/utils", "RobotPrograms_en.json")
GENERAL_YAML_DIR = os.path.join(c.PROJECT_ROOT_DIR, "gherkan/utils", "en-cz_phrases.yaml")

class FillActions:
    def __init__(self):
        self.finetune = Finetuner()

    def ParseMainAction(self, data):
        """ Construct basic NL phrase from given template
            Input format of data:
            data = [Name, VarsVals, Type, Values, TempStr]
            Name = string; name of the template
            VarsVals = list of strings; specific values to fill into template
            Type = "edge"/"bool"/"force"/"equality"
            Values = list of integers; values for given Types
            TempStr = string; specific details for given template
        """
        actor = ""
        self.data = self.convert_to_dic(data)
        lang = self.data["language"]
        values = self.data["vars"]
        inclinations = self.data["inclinations"]

        ## We suppose that templates with robot actions have PN in their name
        if "PN" in self.data["tempName"]:
                phrase = self.data["tempToNL"]
                string = str(self.data['vars']['subject'][1:])  # we leave out the first letter in case it is upper
                subject = (re.findall('([A-Z0-9].*)', string)) # separate actor specification (if exists), i.e. XY
                if subject:
                    subject = subject[0]
                action = self.MatchAction(lang, subject, self.data['vars']['value'])
                verb = self.finetune.find_verb(lang, action)
                if verb:
                    conj_verb = " " + self.finetune.conjugate(lang, verb[0], 3) + " "  # conjugate verb to third person
                    action = re.sub(verb[0], conj_verb, action)
                if subject:
                    subject = self.data['vars']['subject'].replace(subject[0], " " + subject[0])  # separate actor general name, i.e. shuttle
                    self.data["vars"]["subject"] = subject
                self.data["vars"]["value"] = action
                for key, value in values.items():
                      to_replace = re.compile(r"\(\?P\<{}\>[^\)]+\)".format(key))
                      phrase = re.sub(to_replace, value, phrase)
                phrase = self.finetune.strip_extra_spaces(phrase)
        else:
                phrase = self.data["tempToNL"]
                subject = values["subject"]
                if not subject.isupper():
                    string = str(subject[1:]) # we leave out the first letter in case it is upper
                    actor = (re.findall('([A-Z0-9].*)', string))   # separate actor specification (if exists), i.e. XY
                if actor:
                   actor = actor[0]
                   actor_name = subject.replace(actor, "")  # separate actor general name, i.e. shuttle
                else:
                      actor_name = subject
                      actor = ""
                values["subject"] = " ".join([actor_name,actor])
                for key, value in values.items():
                      idx = list(values).index(key)
                      value = self.MatchGeneral(lang, value)
                      if inclinations[idx] != 0:
                       value = self.finetune.conjugate(lang, value, inclinations[idx])
                      to_replace = re.compile(r"\(\?P\<{}\>[^\)]+\)".format(key))
                      phrase = re.sub(to_replace, value, phrase)
                for word in phrase.split():
                    word_new = self.MatchGeneral(lang, word)  # translate words to czech
                    phrase = re.sub(word, word_new, phrase)
                verb = self.finetune.find_verb(lang, phrase)
                if verb:
                  verb_conj = " " + self.finetune.conjugate(lang, verb[0], 3) + " "  # conjugate verb to third person
                  phrase = re.sub(verb[0], verb_conj, phrase)
                phrase = self.finetune.strip_extra_spaces(phrase)
                phrase = self.finetune.align_adjectives(lang, phrase)


        #TODO equality check and so on put to constants
        if phrase:
            data.__dict__["niceStr"] = phrase
        else:
            data.__dict__["niceStr"] = self.data["niceStr"]
            print(colored("Failed to create NL description for following actions: {}".format(self.data["niceStr"]), "yellow"))


    def ParseBool_Robot(self, lang, data, phrase=None):
        """ Return start or end for given bool value """
        if not phrase:
          phrase = ""
        phrase_spec = ""
        if lang == "en":
            if "start" in data[4].lower():
                if data[3] == "1":
                    phrase_spec = ' '.join(["starts the program", phrase])
                elif data[3] == "0":
                    phrase_spec = ' '.join(["does not start the program", phrase])
            elif "end" in data[4].lower():
                if data[3] == "1":
                    phrase_spec = ' '.join(["finishes program", phrase])
                elif data[3] == "0":
                    phrase_spec = ' '.join(["does not finish the program", phrase])
        elif lang =="cz":
            if "start" in data[4].lower():
                if data[3] == "1":
                    phrase_spec = ' '.join(["začne program", phrase])
                elif data[3] == "0":
                    phrase_spec = ' '.join(["nezačne program", phrase])
            elif "end" in data[4].lower():
                if data[3] == "1":
                    phrase_spec = ' '.join(["skončí program", phrase])
                elif data[3] == "0":
                    phrase_spec = ' '.join(["neskončí program", phrase])

        return phrase_spec

    def convert_to_dic(self, lists):
        data = lists.__dict__
        # return [str(el) if not isinstance(el, list) else self.convert_to_str(el) for el in lists]
        return data

    def strip_extra_spaces(self, text):
        stripped_spaces = re.sub(' +', ' ', text)
        stripped_text = stripped_spaces.strip()
        return stripped_text

    def ParseNegative(self, phrase):
        eng_phrase = phrase[0]
        czech_phrase = phrase[1]

        czech_phrase.replace("skončí", "neskončí")
        czech_phrase.replace("začne", "nezačne")
        eng_phrase.replace("finishes", "does not finish")
        eng_phrase.replace("starts", "does not start")

        return self.eng_phrase, self. czech_phrase

    def MatchAction(self, lang, actor, prog):
        """ Load robot actions and return sring for given program number """
        if lang == "en":
            with open(EN_PROGRAM_DICT_DIR, 'r') as stream:
                action_list = json.load(stream)
                try:
                  action = action_list[actor.lower()][str(prog)]
                except:
                  action = "execute program {}".format(prog)
                  print(colored("Warning: Did not find program {} for robot {} in English".format(prog, actor), "yellow"))

        elif lang == "cs":
            with open(CZ_PROGRAM_DICT_DIR, 'r') as stream:
                action_list = json.load(stream)
                try:
                   action = action_list[actor.lower()][str(prog)]
                except:
                  action = "vykonat program {}".format(prog)
                  print(colored("Warning: Did not find program {} for robot {} in Czech".format(prog, actor), "yellow"))

        return action

    def MatchGeneral(self, lang, word):
        """ Return nice string for basic form of a word """
        with open(GENERAL_YAML_DIR, 'r') as stream:
            phrase_list = yaml.load(stream)
            try:
              phrase = phrase_list[lang][word.lower()]
            except:
                phrase = word
        return phrase



if __name__=="__main__":
    fill_action = FillActions()


import json
import re
import copy

from gherkan.containers.SignalStatementTreeNode import SignalStatementTreeOperandNode, \
    SignalStatementTreeOperatorNode
from gherkan.containers.SignalStatementTree import SignalStatementTree
from gherkan.containers.signal_conditional_statement import SignalCondition, NLCondition
from gherkan.decoder.Parser import Parser

class NLParser(Parser):
    def __init__(self):
        super().__init__()

    # TODO rewrite to something more readable
    def parseStatement(self, statement: str):
        """
        Parses a single conditional statement. Uses the regex token specification (see attributes of this class).

        Parameters
        ----------
        statement : str
            A single conditional statement in the form of a string.

        Returns
        -------
        SignalStatementTree:
            Signal tree containing the root of the extracted tree and the corresponding statement string (input to this function)
        """
        operatorStack = []
        outputStack = []

        dataStack = re.split(r'( AND | OR | A | NEBO )', statement)  # brackets to save also the separator value
        dataStackFull = copy.deepcopy(dataStack)  # saves NL statements in original version not with program numbers

        with open(
                self.ROBOT_PROGRAMS_EN) as f:  # loads json with program numbers and descriptions which were extracted from NL input
            programs = json.load(f)

        for id, ot in enumerate(dataStack):
            # checks if the statement is in robot actions. If yes, then it
            # detects robot programs in json file to exchange the actions for program numbers
            re_robotPrograms = re.compile(r"\brobot\s(?P<name>[a-z]\d{1})\s+(?P<text>.*)")
            cond = re_robotPrograms.search(ot)
            if cond:
                robProg = cond.group("text")
                robName = cond.group("name")
                for phrase in programs[robName]:
                    if programs[robName][phrase] in robProg:
                        dataStack[id] = "robot" + robName.upper() + " program number is " + phrase

        for id, ot in enumerate(dataStack):
            if ot in [' AND ', ' OR ', ' A ', ' NEBO ']:
                ot = ot.replace(' ', '')
                operatorStack.append(ot)
            else:
                cond = NLCondition(dataStackFull[id], dataStackFull[id], ot,
                                   ot)  # (strFullEn, strFullCs, strEng, strCZ)
                outputStack.append(cond)

        while len(operatorStack) > 0:
            outputStack.append(operatorStack.pop())

        # At this point, the statement has been converted to a suffix notation. Now it needs to be converted into a tree
        if len(outputStack) == 1:  # single node tree
            root = SignalStatementTreeOperandNode(outputStack[0])
        elif len(outputStack) > 1:
            outputStack.reverse()
            root = None
            lastOperatorNode = None
            for item in outputStack:
                if type(item) is str:  # item == operator or bracket
                    node = SignalStatementTreeOperatorNode(item)
                    if lastOperatorNode:
                        lastOperatorNode.attachChild(node)
                    elif root:
                        raise Exception("Multiple roots!")
                    lastOperatorNode = node
                else:  # item == operand
                    node = SignalStatementTreeOperandNode(item)
                    if lastOperatorNode is not None and lastOperatorNode.attachChild(node):
                        lastOperatorNode = lastOperatorNode.getIncompleteParent()
                if not root:
                    root = node
        else:
            raise Exception("Statement parsing resulted in an empty tree!\n{}".format(statement))

        signalTree = SignalStatementTree(root, statement)
        return signalTree


import re

from gherkan.containers.SignalStatementTreeNode import SignalStatementTreeOperandNode, \
    SignalStatementTreeOperatorNode
from gherkan.containers.SignalStatementTree import SignalStatementTree
from gherkan.containers.signal_conditional_statement import SignalCondition
from gherkan.decoder.Parser import Parser

import gherkan.utils.constants as c


class SignalParser(Parser):
    def __init__(self):
        super().__init__()

        self.tokens_specs = [  # Regex collection for tokens (operators and operands) that can occur in conditions
            (c.AND, r"&&"),
            (c.OR, r"\|\|"),
            (c.NOT, r"!(?!=)"),
            (c.EQUALITY, r"\w+\s*==\s*\w+"),
            (c.INEQUALITY, r"\w+\s*!=\s*\w+"),
            (c.BOOL, r'(\w+)\s*(?![=,!\(\w]+)'),
            (c.EDGE, r"edge\([\w ,]+\)"),
            (c.FORCE, r"force\([\w ,]+\)"),
            (c.UNFORCE, r"unforce\([\w ,]+\)"),
            (c.LBRACKET, r"(?<!\w)\("),
            (c.RBRACKET, r"\)\s*(?=$|[&\|])"),
            (c.SKIP, r'[ \t]+'),
            (c.MISMATCH, r'.'),
        ]
        self.tok_regex = '|'.join(
            '(?P<%s>%s)' % pair for pair in self.tokens_specs)  # combine regex specifications into one giant regex
        self.operands = frozenset([c.EQUALITY, c.INEQUALITY, c.BOOL, c.EDGE, c.FORCE, c.UNFORCE])
        self.operators = [c.NOT, c.AND, c.OR]  # in the order of precedence/priority from highest to lowest

    def parseStatement(self, statement: str):
        """
        Parses a single conditional statement. Uses the regex token specification (see attributes of this class).

        Parameters
        ----------
        statement : str
            A single conditional statement in the form of a string.

        Returns
        -------
        SignalStatementTree:
            Signal tree containing the root of the extracted tree and the corresponding statement string (input to this function)
        """
        operatorStack = []
        outputStack = []

        # TODO: NOT operator is not handled in case it appears in front of brackets
        for mo in re.finditer(self.tok_regex, statement):
            kind = mo.lastgroup
            value = mo.group()
            if kind in self.operators:
                if len(operatorStack) > 0:
                    topOperator = operatorStack[-1]
                    while len(operatorStack) > 0 and self.operatorAPrecedesB(topOperator, kind) and operatorStack[-1] != c.LBRACKET:
                        outputStack.append(operatorStack.pop())
                        topOperator = operatorStack[-1]
                operatorStack.append(kind)
            elif kind == c.LBRACKET:
                # NOT can be handled here by appling DeMorgan to the entire bracket...
                operatorStack.append(kind)
            elif kind == c.RBRACKET:
                while len(operatorStack) > 0 and operatorStack[-1] != c.LBRACKET:
                    outputStack.append(operatorStack.pop())
                operatorStack.pop()  # pop the left bracket onto the output
                if len(operatorStack) > 0 and operatorStack[-1] == c.NOT:
                    # ...or NOT can be handled here somehow
                    operatorStack.pop()
            elif kind in self.operands:
                if len(operatorStack) > 0 and operatorStack[-1] == c.NOT:
                    negate = True
                    operatorStack.pop()
                else:
                    negate = False
                cond = SignalCondition(value, kind, negate)
                outputStack.append(cond)
            elif kind == c.SKIP:
                continue
            elif kind == c.MISMATCH:
                raise RuntimeError(f'{value!r} unexpected')
        while len(operatorStack) > 0:
            outputStack.append(operatorStack.pop())

        # At this point, the statement has been converted to a suffix notation. Now it needs to be converted into a tree
        if len(outputStack) == 1:  # single node tree
            root = SignalStatementTreeOperandNode(outputStack[0])
        elif len(outputStack) > 1:
            outputStack.reverse()
            root = None
            lastOperatorNode = None
            for item in outputStack:
                if type(item) is str:  # item == operator or bracket
                    node = SignalStatementTreeOperatorNode(item)
                    if lastOperatorNode:
                        lastOperatorNode.attachChild(node)
                    elif root:
                        raise Exception("Multiple roots!")
                    lastOperatorNode = node
                else:  # item == operand
                    node = SignalStatementTreeOperandNode(item)
                    if lastOperatorNode is not None and lastOperatorNode.attachChild(node):
                        lastOperatorNode = lastOperatorNode.getIncompleteParent()
                if not root:
                    root = node
        else:
            raise Exception("Statement parsing resulted in an empty tree!\n{}".format(statement))

        signalTree = SignalStatementTree(root, statement)
        return signalTree

    def operatorAPrecedesB(self, a, b):
        """
        Returns True if "a" has higher priority than "b", returns False otherwise.
        """
        if a in self.operators and b in self.operators:
            return self.operators.index(a) < self.operators.index(b)
        else:
            return None


import re

import pkg_resources
import yaml
import warnings
from gherkan.containers.SignalStatementTree import SignalPhrase
from gherkan.containers.SignalStatementTreeNode import SignalStatementTreeOperatorNode, SignalStatementTreeOperandNode, \
    SignalStatementTreeNode

from enum import Enum
from pprint import pprint as pp
from gherkan.processing.Temp2NL import FillActions
from gherkan.utils import constants as c


class TreeProcessor:
    class Direction(Enum):
        SIGNAL_TO_NL = 0
        NL_TO_SIGNAL = 1

    def __init__(self, language):
        self.templ_dic = None
        self.language = language

    def translate_to_signal(self, node: SignalStatementTreeOperandNode, sp: SignalPhrase, template: dict):
        # take the initial pattern TempStr from the template
        variable = template["TempStr"][0]

        # replace all named patterns by their value
        for group, value in sp.vars.items():
            if value is not None:
                strToReplace = re.compile(r"\(\?P\<{}\>[^\)]+\)".format(group))
                variable = re.sub(strToReplace, value, variable)

        # the result should be a variable name without any whitespaces
        node.data.variable = re.sub(r'\s+', '', variable)

        # TODO agree how to fill in the value (hardcoding "1" here does not make much sense)
        node.data.value = sp.vars["value"] if "value" in sp.vars else "[empty]"


    def translate_to_nl(self, node: SignalStatementTreeOperandNode, signalPhrase: SignalPhrase):
        # call Temp2NL
        fill_action = FillActions()
        fill_action.ParseMainAction(data=signalPhrase)

        node.data.variableNL_full = signalPhrase.niceStr

    def extract_variables(self, node: SignalStatementTreeOperandNode, template: dict, match: re.match):
        # fill in fields in SignalPhrase object which will be sent to Temp2NL for further processing
        sp = SignalPhrase()
        sp.tempName = template["Name"]
        sp.tempStr = template["TempStr"][0]
        sp.tempToNL = template["EnTempToNLP"] if self.language == c.LANG_EN else template["CzTempToNLP"]
        sp.inclinations = template["Inclinations"]
        sp.language = self.language

        # TODO should not take only the first element
        type = template["Type"][0]
        node.data.kind = type

        # find all (named) groups in template patterns
        varStrGroups = re.findall(r'(?<=\<).*?(?=\>)', sp.tempStr)
        varNLGroups = re.findall(r'(?<=\<).*?(?=\>)', sp.tempToNL)

        for group in varNLGroups:
            if group in varStrGroups:
                # fill in the matched value of the named pattern
                sp.vars[group] = match[group]

            elif group == "value":
                # value is usually present in matches from NL, otherwise we use the value saved in the node
                sp.vars[group] = match[group] if group in match.groupdict() else node.value
                sp.valTypes[type] = node.value

        return sp

    def process_node(self, node: SignalStatementTreeOperandNode, dir: Direction):
        template, match = self.match_template(node, dir)

        if match is None:
            warnings.warn("No match found for phrase: " + node.data.string)
            return

        signalPhrase = self.extract_variables(node, template, match)

        if dir == TreeProcessor.Direction.SIGNAL_TO_NL:
            # make the real translation to NL using Temp2NL
            self.translate_to_nl(node, signalPhrase)
        elif dir == TreeProcessor.Direction.NL_TO_SIGNAL:
            self.translate_to_signal(node, signalPhrase, template, match)
        else:
            raise NotImplementedError("Direction {} unknown".format(dir))


    def process_tree(self, node: SignalStatementTreeNode, dir: Direction):
        if type(node) is SignalStatementTreeOperatorNode:
            self.process_tree(node.lchild, dir)
            self.process_tree(node.rchild, dir)

        elif type(node) is SignalStatementTreeOperandNode:
            self.process_node(node, dir)
        else:
            raise NotImplementedError("Unrecognized type of node: {}".format(type(node)))


    def match_template(self, node: SignalStatementTreeOperandNode, dir: Direction):
        # finds a matching template in the YAML file (works for both directions)
        template = None
        match = None

        if dir == TreeProcessor.Direction.NL_TO_SIGNAL:
            str_to_match = node.data.variableNLP
        elif dir == TreeProcessor.Direction.SIGNAL_TO_NL:
            str_to_match = node.data.variable
        else:
            raise NotImplementedError("Direction {} unknown".format(dir))

        for index, template in self.templ_dic.items():
            if dir == TreeProcessor.Direction.NL_TO_SIGNAL:
                template_str = template["TempToNLP"]
            else:
                # TODO we take only first element, we should look for all elements in the list
                template_str = template["TempStr"][0]

            expToMatch = re.compile(r"" + template_str)
            match = expToMatch.search(str_to_match)

            if match:
                break

        return template, match

    def load_templ_dic(self, file):
        # load templates to be matched
        with pkg_resources.resource_stream("gherkan", file) as stream:
            try:
                self.templ_dic = yaml.load(stream)
            except yaml.YAMLError as exc:
                print(exc)


"""
This class aligns natural language phrases into their most default forms and matches them
with niceStr parameters located in templates_dic.yaml. Returns an object (like SignalPhrase) with selected template and
filled as many fields from the dic as possible.
"""
import yaml, os, re
import gherkan.utils.constants as c
from gherkan.utils.word_morpho import Finetuner
TEMPLATES_DIC = os.path.join(c.PROJECT_ROOT_DIR, "gherkan/utils", "templates_dic.yaml")

class NLTempMatcher():
    def __init__(self):
        self.temp_dic = self.open_yaml(TEMPLATES_DIC)
        self.finetune = Finetuner()

    def find_temp(self, phrase):
        phrase = self.unify_synonyms(phrase)

        for step in range(len(phrase.split())):
            phrase = self.tag_text(phrase)
            if self.match_with_temps(phrase) == True:
                break
            else:
                continue


    def open_yaml(self, path):
        with open(path, 'r') as stream:
            temp_dic = yaml.load(stream)
        return temp_dic

    def detect_valtype(self, phrase):
        pass

    def detect_negative(self, phrase):

    def tag_text(self, text):
        pass

    def unify_synonyms(self, text):
        pass

    def match_with_temps(self, phrase):
        re.findall(phrase, self.temp_dic)  # generally

        if successful:
            return True

        if fail:
            return False

















if __name__ == "__main__":
   match_temp = NLTempMatcher()